module matrix_market

   use general_module

   implicit none

! Define the matrix type, here for matrices in matrix-market format

   type matrix
      integer, allocatable           :: ival(:)
      real(kind=rp),    allocatable  :: rval(:)
      complex(kind=cp), allocatable  :: cval(:)
      integer, allocatable           :: indx(:)
      integer, allocatable           :: jndx(:)
      integer                        :: rows, cols, nnz
      character*7                    :: field
      character*10                   :: rep
      character*19                   :: symm
   end type matrix

! Overload * to define the matrix-vector multiplication using the matrix type

   INTERFACE OPERATOR(*)
      module procedure rmatvec, cmatvec
   END INTERFACE

! Define the preconditioner type

   type preconditioner
      real(kind=rp),    allocatable  :: RP(:)
      complex(kind=cp), allocatable  :: CP(:)
      character*6,      which_preconditioner
   end type preconditioner

! Overload / to define the preconditioner operation using the preconditioner type

   INTERFACE OPERATOR(/)
      module procedure rprecon, cprecon
   END INTERFACE


   contains

   subroutine read_mm_matrix( file, A )

      implicit none 
      integer, intent(in)           :: file
      type(matrix), intent(out)     :: A

      integer                       :: cols, rows, nnz, nnzmax 
      integer                       :: Allocate_status     
      integer, allocatable          :: ival(:), indx(:), jndx(:)
      double precision, allocatable :: rval(:)
      complex, allocatable          :: cval(:)
      character*10                  :: rep
      character*7                   :: field
      character*19                  :: symm
  
!  
! Read header to determine which space is needed
      call mminfo(file,rep,field,symm,rows,cols,nnzmax)
!
! Allocate the space
      if ( rep == 'coordinate' ) then
         allocate( indx(nnzmax), jndx(nnzmax), stat = Allocate_status)
      end if
      if ( field == 'real' ) then
         allocate( rval(nnzmax), stat = Allocate_status)
      elseif ( field == 'integer' ) then
         allocate( ival(nnzmax), stat = Allocate_status)
      elseif ( field == 'complex' ) then
         allocate( cval(nnzmax), stat = Allocate_status)
      end if
      call mmread(file,rep,field,symm,rows,cols,nnz,nnzmax, indx,jndx,ival,rval,cval )
!
! Now store everything in the matrix:
      A%rep = rep
      A%field = field
      A%symm = symm
      A%rows = rows
      A%cols = cols
      A%nnz = nnz
      if ( rep == 'coordinate' ) then
         allocate( A%indx(nnzmax), A%jndx(nnzmax), stat = Allocate_status)
         A%indx = indx
         A%jndx = jndx
         deallocate( indx, jndx )
      end if
      if ( field == 'real' ) then
         allocate( A%rval(nnzmax), stat = Allocate_status)
         A%rval = rval
         deallocate( rval )
      elseif ( field == 'integer' ) then
         allocate( A%ival(nnzmax), stat = Allocate_status)
         A%ival = ival
         deallocate( ival )
      elseif ( field == 'complex' ) then
         allocate( A%cval(nnzmax), stat = Allocate_status)
         A%cval = cval
         deallocate( cval )
      end if

   end subroutine read_mm_matrix 

   function rmatvec( A, v )

      type(matrix), intent(in)        :: A
      real(kind=rp), intent(in)       :: v(:,:)
      real(kind=rp)                   :: rmatvec(A%rows,size(v,2)), w(A%rows,size(v,2))
      integer                         :: m, n, nnz, i, j, k

      m   = A%rows
      n   = A%cols
      nnz = A%nnz

! Initialization:
      w = 0.
      if ( A%field == 'real' .and. A%rep == 'array' .and. A%symm == 'general' ) then
!
! General real matrix, dense matrix-vector product:
         k = 0
         do j = 1,m
            do i = 1, n
               k = k + 1
               w(i,:) = w(i,:) + A%rval(k)*v(j,:)
            end do
         end do
!
      elseif ( A%field == 'real' .and. A%rep == 'array' .and. A%symm == 'symmetric' ) then
!
! Symmetric real matrix, dense matrix-vector product:
         k = 0
         do j = 1,m
            do i = j, n
               k = k + 1
               w(i,:) = w(i,:) + A%rval(k)*v(j,:)
               if ( j /= i ) then
                  w(j,:) = w(j,:) + A%rval(k)*v(i,:)
               end if
            end do
         end do
!
      elseif ( A%field == 'real' .and. A%rep == 'array' .and. A%symm == 'skew-symmetric' ) then
!
! Skew-symmetric real matrix, dense matrix-vector product:
         k = 0
         do j = 1,m
            do i = j+1, n
               k = k + 1
               w(i,:) = w(i,:) + A%rval(k)*v(j,:)
               w(j,:) = w(j,:) - A%rval(k)*v(i,:)
            end do
         end do
!
      elseif ( A%field == 'real' .and. A%rep == 'coordinate' .and. A%symm == 'general' ) then
!
! General real matrix, sparse matrix-vector product:
         do j = 1,nnz
            w(A%indx(j),:) = w(A%indx(j),:) + A%rval(j)*v(A%jndx(j),:)
         end do
!
      elseif ( A%field == 'real' .and. A%rep == 'coordinate' .and. A%symm == 'symmetric' ) then
!
! Symmetric real matrix, sparse matrix-vector product:
         do j = 1,nnz
            w(A%indx(j),:) = w(A%indx(j),:) + A%rval(j)*v(A%jndx(j),:)
            if ( A%indx(j) /= A%jndx(j) ) then
               w(A%jndx(j),:) = w(A%jndx(j),:) + A%rval(j)*v(A%indx(j),:)
            end if
         end do
!
      elseif ( A%field == 'real' .and. A%rep == 'coordinate' .and. A%symm == 'skew-symmetric' ) then
!
! Skew-symmetric real matrix, sparse matrix-vector product:
         do j = 1,nnz
            w(A%indx(j),:) = w(A%indx(j),:) + A%rval(j)*v(A%jndx(j),:)
            w(A%jndx(j),:) = w(A%jndx(j),:) - A%rval(j)*v(A%indx(j),:)
         end do
      elseif ( A%field == 'integer' .and. A%rep == 'array' .and. A%symm == 'general' ) then
!
! General integer matrix, dense matrix-vector product:
         k = 0
         do j = 1,m
            do i = 1, n
               k = k + 1
               w(i,:) = w(i,:) + A%ival(k)*v(j,:)
            end do
         end do
!
      elseif ( A%field == 'integer' .and. A%rep == 'array' .and. A%symm == 'symmetric' ) then
!
! Symmetric integer matrix, dense matrix-vector product:
         k = 0
         do j = 1,m
            do i = j, n
               k = k + 1
               w(i,:) = w(i,:) + A%ival(k)*v(j,:)
               if ( j /= i ) then
                  w(j,:) = w(j,:) + A%ival(k)*v(i,:)
               end if
            end do
         end do
!
      elseif ( A%field == 'integer' .and. A%rep == 'array' .and. A%symm == 'skew-symmetric' ) then
!
! Skew-symmetric integer matrix, Dense matrix-vector product:
         k = 0
         do j = 1,m
            do i = j+1, n
               k = k + 1
               w(i,:) = w(i,:) + A%ival(k)*v(j,:)
               w(j,:) = w(j,:) - A%ival(k)*v(i,:)
            end do
         end do
!
      elseif ( A%field == 'integer' .and. A%rep == 'coordinate' .and. A%symm == 'general' ) then
!
! General integer matrix, sparse matrix-vector product
         do j = 1,nnz
            w(A%indx(j),:) = w(A%indx(j),:) + A%ival(j)*v(A%jndx(j),:)
         end do
!
      elseif ( A%field == 'integer' .and. A%rep == 'coordinate' .and. A%symm == 'symmetric' ) then
!
! Symmetric integer matrix, sparse matrix-vector product
         do j = 1,nnz
            w(A%indx(j),:) = w(A%indx(j),:) + A%ival(j)*v(A%jndx(j),:)
            if ( A%indx(j) /= A%jndx(j) ) then
               w(A%jndx(j),:) = w(A%jndx(j),:) + A%ival(j)*v(A%indx(j),:)
            end if
         end do
!
      elseif ( A%field == 'integer' .and. A%rep == 'coordinate' .and. A%symm == 'skew-symmetric' ) then
!
! Skew-symmetric integer matrix, sparse matrix-vector product
         do j = 1,nnz
            w(A%indx(j),:) = w(A%indx(j),:) + A%rval(j)*v(A%jndx(j),:)
            w(A%jndx(j),:) = w(A%jndx(j),:) - A%rval(j)*v(A%indx(j),:)
         end do
!
      elseif ( A%field == 'pattern' .and. A%rep == 'array' ) then
!
! Binary matrices cannot be dense
!
         stop 'pattern and array incompatible'
!
      elseif ( A%field == 'pattern' .and. A%rep == 'coordinate' .and. A%symm == 'general' ) then
!
! General binary matrix, sparse matrix-vector product
         w = 0.
         do j = 1,nnz
            w(A%indx(j),:) = w(A%indx(j),:) + v(A%jndx(j),:)
         end do
!
      elseif ( A%field == 'pattern' .and. A%rep == 'coordinate' .and. A%symm == 'symmetric' ) then
!
! Symmetric binary matrix, sparse matrix-vector product
         do j = 1,nnz
            w(A%indx(j),:) = w(A%indx(j),:) + v(A%jndx(j),:)
            if ( A%indx(j) /= A%jndx(j) ) then
               w(A%jndx(j),:) = w(A%jndx(j),:) + v(A%indx(j),:)
            end if
         end do
!
      elseif ( A%field == 'pattern' .and. A%rep == 'coordinate' .and. A%symm == 'skew-symmetric' ) then
!
! Skew-symmetric binary matrix, sparse matrix-vector product
         do j = 1,nnz
            w(A%indx(j),:) = w(A%indx(j),:) + v(A%jndx(j),:)
            w(A%jndx(j),:) = w(A%jndx(j),:) - v(A%indx(j),:)
         end do
!
      end if
      rmatvec = w

   end function rmatvec

   function cmatvec( A, v ) 

      type(matrix), intent(in)        :: A
      complex(kind=cp), intent(in)    :: v(:,:)
      complex(kind=cp)                :: cmatvec(A%rows,size(v,2)), w(A%cols,size(v,2))
      integer                         :: m, n, nnz, i, j, k

      m   = A%rows
      n   = A%cols
      nnz = A%nnz

! Initialization:
      w = 0.
      if ( A%rep == 'array' .and. A%symm == 'general' ) then
!
! General complex matrix, dense matrix-vector product:
         k = 0
         do j = 1,n
            do i = 1, m
               k = k +1
               w(i,:) = w(i,:) + A%cval(k)*v(j,:)
            end do
         end do
      elseif ( A%rep == 'array' .and. A%symm == 'symmetric' ) then
!
! Symmetric complex matrix, dense matrix-vector product:
         k = 0
         do j = 1,n
            do i = j,m
               k = k +1
               w(i,:) = w(i,:) + A%cval(k)*v(j,:)
               if ( j /= i ) then
                  w(j,:) = w(j,:) + A%cval(k)*v(i,:)
               end if
            end do
         end do
!
      elseif ( A%rep == 'array' .and. A%symm == 'skew-symmetric' ) then
!
! Skew-symmetric complex matrix, dense matrix-vector product:
         k = 0
         do j = 1,n
            do i = j+1,m
               k = k +1
               w(i,:) = w(i,:) + A%cval(k)*v(j,:)
               w(j,:) = w(j,:) - A%cval(k)*v(i,:)
            end do
         end do
!
      elseif ( A%rep == 'array' .and. A%symm == 'hermitian' ) then
!
! Hermitian complex matrix, dense matrix-vector product:
         k = 0
         do j = 1,n
            do i = j,m
               k = k +1
               w(i,:) = w(i,:) + A%cval(k)*v(j,:)
               if ( j /= i ) then
                  w(j,:) = w(j,:) + conjg(A%cval(k))*v(i,:)
               end if
            end do
         end do
!
      elseif ( A%rep == 'coordinate' .and. A%symm == 'general' ) then
!
! General complex matrix, sparse matrix-vector product:
         do j = 1,nnz
            w(A%indx(j),:) = w(A%indx(j),:) + A%cval(j)*v(A%jndx(j),:)
         end do
!
      elseif ( A%rep == 'coordinate' .and. A%symm == 'symmetric' ) then
!
! Symmetric complex matrix, sparse matrix-vector product:
         do j = 1,nnz
            w(A%indx(j),:) = w(A%indx(j),:) + A%cval(j)*v(A%jndx(j),:)
            if ( A%indx(j) /= A%jndx(j) ) then
               w(A%jndx(j),:) = w(A%jndx(j),:) + A%cval(j)*v(A%indx(j),:)
            end if
         end do
!
      elseif ( A%rep == 'coordinate' .and. A%symm == 'skew_symmetric' ) then
!
! Skew-symmetric complex matrix, sparse matrix-vector product:
         do j = 1,nnz
            w(A%indx(j),:) = w(A%indx(j),:) + A%cval(j)*v(A%jndx(j),:)
            w(A%jndx(j),:) = w(A%jndx(j),:) - A%cval(j)*v(A%indx(j),:)
         end do
!
      elseif ( A%rep == 'coordinate' .and. A%symm == 'hermitian' ) then
!
! Hermitian complex matrix, sparse matrix-vector product:
         do j = 1,nnz
            w(A%indx(j),:) = w(A%indx(j),:) + A%cval(j)*v(A%jndx(j),:)
            if ( A%indx(j) /= A%jndx(j) ) then
               w(A%jndx(j),:) = w(A%jndx(j),:) + conjg(A%cval(j))*v(A%indx(j),:)
            end if
         end do
      end if
      cmatvec = w

   end function cmatvec

   function rprecon( v, M )

      type(preconditioner), intent(in)      :: M
      real(kind=rp), intent(in)             :: v(:,:)
      real(kind=rp)                         :: rprecon(size(v,1),size(v,2)), w(size(v,1),size(v,2))
      integer                               :: nrhs, i

      nrhs = size(v,2)

      if ( M%which_preconditioner == 'none' ) then
         w = v
      elseif ( M%which_preconditioner == 'jacobi' ) then
         do i = 1,nrhs
            w(:,i) = v(:,i)/M%RP
         end do
      end if
      rprecon = w

   end function rprecon

   function cprecon( v, M )

      type(preconditioner), intent(in)      :: M
      complex(kind=cp), intent(in)          :: v(:,:)
      complex(kind=cp)                      :: cprecon(size(v,1),size(v,2)), w(size(v,1),size(v,2))
      integer                               :: nrhs, i

      nrhs = size(v,2)

      if ( M%which_preconditioner == 'none' ) then
         w = v
      else if ( M%which_preconditioner == 'jacobi' ) then
         do i = 1,nrhs
            w(:,i) = v(:,i)/M%CP
         end do
      end if
      cprecon = w

   end function cprecon

   subroutine make_preconditioner( A, M )

      integer                               :: prec
      type(matrix), intent(in)              :: A
      type(preconditioner), intent(out)     :: M
      integer                               :: neq, nnz, i, j, k, err

      neq = A%cols
      nnz = A%nnz
      
      if ( A%symm == 'skew-symmetric' .or. A%field == 'pattern' .or. A%rep == 'array' ) then
         M%which_preconditioner = 'none'
      else
         M%which_preconditioner = 'jacobi'
         if ( A%field == 'complex' ) then
            allocate( M%CP(neq), stat = err ) 
         else
            allocate( M%RP(neq), stat = err ) 
         endif

         if  ( A%field == 'real' ) then
            M%RP = 0.
            do k = 1,nnz
               i = A%indx(k)
               j = A%jndx(k)
               if ( i == j ) then
                  M%RP(i) = M%RP(i) + A%rval(k)
               end if
            end do
         elseif  ( A%field == 'integer' ) then
            M%RP = 0.
            do k = 1,nnz
               i = A%indx(k)
               j = A%jndx(k)
               if ( i == j ) then
                  M%RP(i) = M%RP(i) + A%ival(k)
               end if
            end do
         elseif  ( A%field == 'complex' ) then
            M%CP = 0.
            do k = 1,nnz
               i = A%indx(k)
               j = A%jndx(k)
               if ( i == j ) then
                  M%CP(i) = M%CP(i) + A%cval(k)
               end if
            end do
         end if

! Check if there are zero elements on the main diagonal:
!
         if ( A%field == 'complex' ) then
            do i = 1, neq
               if ( M%CP(i) == 0. ) then
                  M%which_preconditioner = 'none'
                  exit
               end if
            end do 
         else
            do i = 1, neq
               if ( M%RP(i) == 0. ) then
                  M%which_preconditioner = 'none'
                  exit
               end if
            end do 
         end if
                  
      end if

   end subroutine make_preconditioner

end module matrix_market
